package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.GlobalDefaults
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.defs.BaseMovementParameters
import ru.dbotthepony.kstarbound.math.roundTowardsPositiveInfinity
import ru.dbotthepony.kstarbound.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.physics.CollisionPoly
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import java.util.EnumSet
import kotlin.concurrent.withLock
import kotlin.math.absoluteValue

abstract class Entity(val world: World<*, *>) {
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (isRemoved) {
				throw IllegalStateException("This entity was removed")
			} else if (value == field) {
				return
			}

			val chunkPos = world.chunkFromCell(position)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to (expected chunk $chunkPos, got chunk ${value.pos})")
			}

			val oldChunk = field
			field = value

			world.entityListsLock.withLock {
				if (oldChunk == null && value != null) {
					world.orphanedEntities.remove(this)
					value.addEntity(this)
				} else if (oldChunk != null && value == null) {
					world.orphanedEntities.add(this)
					oldChunk.removeEntity(this)
				} else if (oldChunk != null && value != null) {
					value.transferEntity(this, oldChunk)
				}
			}
		}

	var position = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = Vector2d(world.x.cell(value.x), world.y.cell(value.y))
			physicsSleepTicks = 0

			if (isSpawned && !isRemoved) {
				val oldChunkPos = world.chunkFromCell(old)
				val newChunkPos = world.chunkFromCell(field)

				if (oldChunkPos != newChunkPos) {
					chunk = world.chunkMap[newChunkPos]
				}
			}
		}

	var velocity = Vector2d.ZERO
		set(value) {
			field = value
			physicsSleepTicks = 0
		}

	var physicsSleepTicks = 0
	val mailbox = MailboxExecutorService(world.mailbox.thread)

	protected val hitboxes = ArrayList<Poly>()
	protected val collisionFilter: EnumSet<CollisionType> = EnumSet.of(CollisionType.NONE)

	/**
	 * true - whitelist, false - blacklist
	 */
	protected var collisionFilterMode = false

	open var movementParameters: BaseMovementParameters = GlobalDefaults.movementParameters

	/**
	 * Whenever is this entity spawned in world ([spawn] called).
	 * Doesn't mean entity still exists in world, check it with [isRemoved]
	 */
	var isSpawned = false
		private set

	/**
	 * Whenever is this entity was removed from world ([remove] called).
	 */
	var isRemoved = false
		private set

	open fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world.chunkMap[world.chunkFromCell(position)]

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}
	}

	open fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true
		mailbox.shutdownNow()

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}
	}

	/**
	 * this function is executed sequentially
	 */
	open fun thinkShared() {
		mailbox.executeQueuedTasks()
	}

	open fun thinkClient() {

	}

	open fun thinkServer() {

	}

	/**
	 * this function is executed in parallel
	 */
	// TODO: Ghost collisions occur, where objects trip on edges
	open fun move() {
		if (physicsSleepTicks > PHYSICS_TICKS_UNTIL_SLEEP) return
		var physicsSleepTicks = physicsSleepTicks
		velocity += world.gravity * Starbound.TICK_TIME_ADVANCE

		movementParameters.speedLimit.ifPresent {
			if (velocity.length > it) {
				velocity = velocity.unitVector * it
			}
		}

		if (hitboxes.isEmpty()) {
			position += velocity * Starbound.TICK_TIME_ADVANCE
			return
		}

		val steps = roundTowardsPositiveInfinity(velocity.length / 30.0 / hitboxes.stream().map { it.aabb }.reduce(AABB::combine).get().let { it.width.coerceAtLeast(it.height).coerceAtLeast(0.1) })
		val dt = Starbound.TICK_TIME_ADVANCE / steps

		for (step in 0 until steps) {
			position += velocity * dt

			for (i in 0 until 10) {
				val localHitboxes = hitboxes.map { it + position }

				val polies = world.queryCollisions(
					localHitboxes.stream().map { it.aabb }.reduce(AABB::combine).get().enlarge(2.0, 2.0)
				).filter {
					if (collisionFilterMode)
						it.type in collisionFilter
					else
						it.type !in collisionFilter
				}

				if (polies.isEmpty()) break

				val intersects = ArrayList<Poly.Penetration<CollisionPoly>>()

				localHitboxes.forEach { hitbox ->
					polies.forEach { poly -> hitbox.intersect(poly.poly, poly)?.let { intersects.add(it) } }
				}

				if (intersects.isEmpty()) {
					break
				} else {
					val max = intersects.max()
					// resolve collision
					position += max.vector
					// collision response
					val response = max.axis * velocity.dot(max.axis * (1.0 + max.data.bounceFactor) * (1.0 + movementParameters.bounceFactor.orElse(0.0)))
					velocity -= response

					val gravityDot = world.gravity.unitVector.dot(max.axis)
					// impulse?
					velocity += max.data.velocity * gravityDot * dt
					// friction
					velocity *= 1.0 - gravityDot.absoluteValue * 0.08

					onTouch(response, max.axis, max.data)
				}
			}
		}

		if (velocity.lengthSquared < 0.25) {
			physicsSleepTicks++
		}

		this.physicsSleepTicks = physicsSleepTicks
	}

	protected open fun onTouch(velocity: Vector2d, normal: Vector2d, poly: CollisionPoly) {

	}

	open fun render(client: StarboundClient = StarboundClient.current()) {
		hitboxes.forEach { (it + position).render(client) }

		world.queryCollisions(
			hitboxes.stream().map { (it + position).aabb }.reduce(AABB::combine).get().enlarge(2.0, 2.0)
		).filter {
			if (collisionFilterMode)
				it.type in collisionFilter
			else
				it.type !in collisionFilter
		}.forEach { it.poly.render(client, BLOCK_COLLISION_COLOR) }
	}

	open var maxHealth = 0.0
	open var health = 0.0

	open fun hurt(amount: Double): Boolean {
		return false
	}

	companion object {
		const val PHYSICS_TICKS_UNTIL_SLEEP = 16
		val BLOCK_COLLISION_COLOR = RGBAColor(65, 179, 217)
	}
}
