package ru.dbotthepony.kstarbound.world.physics

import com.google.common.collect.ImmutableList
import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.lwjgl.opengl.GL11.GL_LINES
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kstarbound.io.json.consumeNull
import ru.dbotthepony.kstarbound.io.json.listAdapter
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.arrays.Matrix3f
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.intersectSegments
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import kotlin.math.absoluteValue

private fun calculateEdges(points: List<Vector2d>): ImmutableList<Poly.Edge> {
	require(points.size >= 2) { "Provided poly is invalid (only ${points.size} points are defined)" }

	val edges = ImmutableList.Builder<Poly.Edge>()

	for (i in points.indices) {
		val p0 = points[i]
		val p1 = points[(i + 1) % points.size]

		val diff = (p1 - p0).unitVector
		val normal = Vector2d(-diff.y, diff.x)

		edges.add(Poly.Edge(p0, p1, normal))
	}

	return edges.build()
}

/**
 * edges are built in clockwise winding
 *
 * If poly shape is not convex behavior of SAT is undefined
 */
class Poly private constructor(val edges: ImmutableList<Edge>, val vertices: ImmutableList<Vector2d>) {
	constructor(points: List<Vector2d>) : this(calculateEdges(points), ImmutableList.copyOf(points))
	constructor(aabb: AABB) : this(listOf(aabb.bottomLeft, aabb.topLeft, aabb.topRight, aabb.bottomRight))

	val aabb = AABB(
		Vector2d(vertices.minOf { it.x }, vertices.minOf { it.y }),
		Vector2d(vertices.maxOf { it.x }, vertices.maxOf { it.y }),
	)

	data class Edge(val p0: Vector2d, val p1: Vector2d, val normal: Vector2d) {
		operator fun plus(other: IStruct2d): Edge {
			return Edge(p0 + other, p1 + other, normal)
		}

		operator fun minus(other: IStruct2d): Edge {
			return Edge(p0 - other, p1 - other, normal)
		}

		operator fun times(other: IStruct2d): Edge {
			return Edge(p0 * other, p1 * other, (normal * other).unitVector)
		}

		operator fun times(other: Double): Edge {
			return Edge(p0 * other, p1 * other, (normal * other).unitVector)
		}
	}

	data class Penetration<T>(val axis: Vector2d, val penetration: Double, val data: T) : Comparable<Penetration<*>> {
		val vector = axis * penetration

		override fun compareTo(other: Penetration<*>): Int {
			return penetration.absoluteValue.compareTo(other.penetration.absoluteValue)
		}
	}

	operator fun plus(value: Penetration<*>): Poly {
		val vertices = ImmutableList.Builder<Vector2d>()
		val edges = ImmutableList.Builder<Edge>()

		for (v in this.vertices) vertices.add(v + value.vector)
		for (v in this.edges) edges.add(v + value.vector)

		return Poly(edges.build(), vertices.build())
	}

	operator fun plus(value: IStruct2d): Poly {
		val vertices = ImmutableList.Builder<Vector2d>()
		val edges = ImmutableList.Builder<Edge>()

		for (v in this.vertices) vertices.add(v + value)
		for (v in this.edges) edges.add(v + value)

		return Poly(edges.build(), vertices.build())
	}

	operator fun minus(value: IStruct2d): Poly {
		val vertices = ImmutableList.Builder<Vector2d>()
		val edges = ImmutableList.Builder<Edge>()

		for (v in this.vertices) vertices.add(v - value)
		for (v in this.edges) edges.add(v - value)

		return Poly(edges.build(), vertices.build())
	}

	operator fun times(value: IStruct2d): Poly {
		val vertices = ImmutableList.Builder<Vector2d>()
		val edges = ImmutableList.Builder<Edge>()

		for (v in this.vertices) vertices.add(v * value)
		for (v in this.edges) edges.add(v * value)

		return Poly(edges.build(), vertices.build())
	}

	operator fun times(value: Double): Poly {
		val vertices = ImmutableList.Builder<Vector2d>()
		val edges = ImmutableList.Builder<Edge>()

		for (v in this.vertices) vertices.add(v * value)
		for (v in this.edges) edges.add(v * value)

		return Poly(edges.build(), vertices.build())
	}

	// min / max
	fun project(normal: Vector2d): IStruct2d {
		var min = vertices.first().dot(normal)
		var max = min

		for (vertex in vertices) {
			val dist = vertex.dot(normal)

			min = min.coerceAtMost(dist)
			max = max.coerceAtLeast(dist)
		}

		return Vector2d(min, max)
	}

	fun <T> intersect(other: Poly, data: T): Penetration<T>? {
		if (!aabb.intersectWeak(other.aabb))
			return null

		val normals = ObjectOpenHashSet<Vector2d>()
		edges.forEach { normals.add(it.normal) }
		other.edges.forEach { normals.add(it.normal) }

		val intersections = ArrayList<Penetration<T>>()

		for (normal in normals) {
			val projectThis = project(normal)
			val projectOther = other.project(normal)

			if (!intersectSegments(projectThis.component1(), projectThis.component2(), projectOther.component1(), projectOther.component2())) {
				return null
			} else {
				val width = projectThis.component2() - projectThis.component1()

				if (
					projectOther.component1() in projectThis.component1() .. projectThis.component2() &&
					projectOther.component2() in projectThis.component1() .. projectThis.component2()
				) {
					// other inside this
					val minMin = projectThis.component1() - projectOther.component1()
					val maxMax = projectThis.component2() - projectOther.component2()

					if (minMin <= maxMax) {
						// push to left
						intersections.add(Penetration(normal, -minMin - width, data))
					} else {
						// push to right
						intersections.add(Penetration(normal, maxMax + width, data))
					}
				} else if (
					projectThis.component1() in projectOther.component1() .. projectOther.component2() &&
					projectThis.component2() in projectOther.component1() .. projectOther.component2()
				) {
					// this inside other
					val minMin = projectThis.component1() - projectOther.component1()
					val maxMax = projectOther.component2() - projectThis.component2()

					if (minMin <= maxMax) {
						// push to left
						intersections.add(Penetration(normal, -minMin - width, data))
					} else {
						// push to right
						intersections.add(Penetration(normal, maxMax + width, data))
					}
				} else if (projectOther.component1() in projectThis.component1() .. projectThis.component2()) {
					// other's min point is within this
					intersections.add(Penetration(normal, projectOther.component1() - projectThis.component2(), data))
				} else {
					// other's max point in within this
					intersections.add(Penetration(normal, projectOther.component2() - projectThis.component1(), data))
				}

				if (intersections.last().penetration == 0.0) {
					return null
				}
			}
		}

		if (intersections.isEmpty())
			return null

		return intersections.min()
	}

	fun intersect(other: Poly) = intersect(other, Unit)

	fun render(client: StarboundClient = StarboundClient.current(), color: RGBAColor = RGBAColor.LIGHT_GREEN) {
		val program = client.programs.position
		val lines = program.builder.builder
		program.use()
		lines.begin(GeometryType.LINES)

		val normalSize = (aabb.width.coerceAtMost(aabb.height) / 3.0).coerceAtLeast(0.1)

		for (edge in edges) {
			val current = edge.p0
			val next = edge.p1
			lines.vertex(client.stack.last(), current)
			lines.vertex(client.stack.last(), next)
			lines.vertex(client.stack.last(), (next + current) / 2.0)
			lines.vertex(client.stack.last(), (next + current) / 2.0 + edge.normal * normalSize)
		}

		program.modelMatrix = identity
		program.colorMultiplier = color
		program.builder.upload()
		program.builder.draw(GL_LINES)
	}

	override fun toString(): String {
		return "Poly[aabb = $aabb; edges = $edges]"
	}

	override fun equals(other: Any?): Boolean {
		return other === this || other is Poly && vertices == other.vertices
	}

	override fun hashCode(): Int {
		return vertices.hashCode()
	}

	companion object : TypeAdapterFactory {
		private val identity = Matrix3f.identity()

		override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (type.rawType === Poly::class.java) {
				return object : TypeAdapter<Poly>() {
					private val list = gson.listAdapter<Vector2d>()

					override fun write(out: JsonWriter, value: Poly?) {
						list.write(out, value?.vertices)
					}

					override fun read(`in`: JsonReader): Poly? {
						if (`in`.consumeNull())
							return null
						else
							return Poly(list.read(`in`))
					}
				} as TypeAdapter<T>
			}

			return null
		}
	}
}
